@@ -5,47 +5,21 @@
 package HTML::Mason::Commands;
 
 use strict;
+use Date::Manip;
 use File::Basename;
 use HTML::Mason::Utils;
 use HTML::Mason::Tools qw(read_file chop_slash);
 use HTML::Mason::Config;
 use IO;
-use Time::Local;
+use URI::Escape;
 
-use vars qw($INTERP @ISA @EXPORT_OK @EXPORT);
- 
-require Exporter;
-@ISA=qw(Exporter);
-@EXPORT=qw(
-	mc_abort
-        mc_auto_comp
-	mc_auto_next
-	mc_cache
-	mc_cache_self
-	mc_caller 
-	mc_comp
-	mc_comp_exists
-	mc_comp_source
-	mc_comp_stack
-	mc_date 
-	mc_file 
-	mc_file_root 
-	mc_filter_self
-	mc_out 
-	mc_time
-);
-
-@EXPORT_OK=@EXPORT;
+use vars qw($INTERP);
 
 #
 # Convert relative paths to absolute, handle . and ..
-# Empty string resolves to current component path.
 #
 my $process_comp_path = sub {
     my ($compPath) = @_;
-    if ($compPath !~ /\S/) {
-	return $INTERP->locals->{callPath};
-    }
     if ($compPath !~ m@^/@) {
 	$compPath = chop_slash($INTERP->locals->{parentPath}) . "/" . $compPath;
     }
@@ -54,45 +28,42 @@ my $process_comp_path = sub {
     return $compPath;
 };
 
-my $no_interp_error = "called outside of Interp::exec environment";
-my $no_auto_error = "called when no autohandler invoked";
+#
+# Construct a CGI-style option string from a hash.
+#
+my $construct_option_string = sub {
+    my (%options) = @_;
+    my ($key,$value,$ostr);
+    return '' if (!%options);
+    while (($key,$value) = each(%options)) {
+	if (!ref($value)) {
+	    $ostr .= "$key=".uri_escape($value)."&";
+	} elsif (ref($value) eq 'ARRAY') {
+	    foreach (@$value) {
+		$ostr .= "$key=".uri_escape($_)."&";
+	    }
+	} elsif (ref($value) eq 'HASH') {
+	    my %h = %$value;
+	    foreach (keys(%h)) {
+		$ostr .= "$key=".uri_escape($_)."&$key=".uri_escape($h{$_})."&";
+	    }
+	} else {
+	    die "cannot pass ".ref($value)." reference in option string\n";
+	}
+    }
+    chop($ostr);
+    return $ostr;
+};
 
 sub mc_abort
 {
-    die "mc_abort $no_interp_error" if !$INTERP;
-    $INTERP->abort(@_);
-}
-
-sub mc_auto_comp
-{
-    die "mc_auto_comp $no_interp_error" if !$INTERP;
-    my $aref = $INTERP->{exec_state}->{autohandler_next} or die "mc_auto_comp $no_auto_error";
-    my $path = $aref->[0];
-    
-    # return relative path if possible
-    my $curdir = $INTERP->locals->{parentPath};
-    $path =~ s{^$curdir/}{};
-    return $path;
-}
-
-sub mc_auto_next
-{
-    die "mc_auto_next $no_interp_error" if !$INTERP;
-    my $aref = $INTERP->{exec_state}->{autohandler_next} or die "mc_auto_next $no_auto_error";
-    my ($compPath, $argsref) = @$aref;
-    my %args = (%$argsref,@_);
-    my ($result,@result);
-    if (wantarray) {
-	@result = mc_comp($compPath, %args);
-    } else {
-	$result = mc_comp($compPath, %args);
-    }
-    return wantarray ? @result : $result;
+    $INTERP->{exec_state}->{abort_flag} = 1;
+    $INTERP->{exec_state}->{abort_retval} = $_[0];
+    die "aborted";
 }
 
 sub mc_cache
 {
-    die "mc_cache $no_interp_error" if !$INTERP;
     my (%options) = @_;
     return undef if !$INTERP->use_data_cache;
     $options{cache_file} = $INTERP->data_cache_filename($INTERP->locals->{truePath});
@@ -100,28 +71,17 @@ sub mc_cache
 	$options{memory_cache} = $INTERP->{data_cache_store};
 	delete($options{keep_in_memory});
     }
-    
-    $options{action} = $options{action} || 'retrieve';
-    $options{key} = $options{key} || 'main';
-    my $results = HTML::Mason::Utils::access_data_cache(%options);
-    if ($options{action} eq 'retrieve') {
-	$INTERP->write_system_log('CACHE_READ',$INTERP->locals->{truePath},$options{key},
-				  defined $results ? 1 : 0);
-    } elsif ($options{action} eq 'store') {
-	$INTERP->write_system_log('CACHE_WRITE',$INTERP->locals->{truePath},$options{key});
-    }
-    return $results;
+    return HTML::Mason::Utils::access_data_cache(%options);
 }
 
 sub mc_cache_self
 {
-    die "mc_cache_self $no_interp_error" if !$INTERP;
     my (%options) = @_;
     
     return 0 if !$INTERP->use_data_cache;
     return 0 if $INTERP->locals->{inCacheSelfFlag};
     my (%retrieveOptions,%storeOptions);
-    foreach (qw(key expire_if keep_in_memory busy_lock)) {
+    foreach (qw(key expire_if keep_in_memory)) {
 	if (exists($options{$_})) {
 	    $retrieveOptions{$_} = $options{$_};
 	}
@@ -147,8 +107,11 @@ sub mc_cache_self
 	$INTERP->{stack}->[0] = {%saveLocals};
 	mc_cache(action=>'store',value=>$result,%storeOptions);
     } else {
-	$INTERP->call_hooks('start_primary');
-	$INTERP->call_hooks('end_primary');
+	#
+	# Hack! http header is technically a side-effect, so we must
+	# call it explicitly.
+	#
+	$INTERP->call_hooks(name=>'http_header') if ($INTERP->depth==1);
     }
     mc_out($result);
     return 1;
@@ -156,7 +119,6 @@ sub mc_cache_self
 
 sub mc_caller ()
 {
-    die "mc_caller $no_interp_error" if !$INTERP;
     if ($INTERP->depth <= 1) {
 	return undef;
     } else {
@@ -164,45 +126,13 @@ sub mc_caller ()
     }
 }
 
-sub mc_call_self
-{
-    die "mc_call_self $no_interp_error" if !$INTERP;
-    my ($cref,$rref) = @_;
-    return 0 if $INTERP->locals->{inCallSelfFlag};
-    
-    #
-    # Reinvoke the component with inCallSelfFlag=1. Collect
-    # output and return value in references provided.
-    #
-    my $content;
-    my $lref = $INTERP->{stack}->[0];
-    my %saveLocals = %$lref;
-    $lref->{sink} = sub { $content .= $_[0] };
-    $lref->{inCallSelfFlag} = 1;
-    my $sub = $lref->{callFunc};
-    my %args = %{$lref->{callArgs}};
-    if (ref($rref) eq 'SCALAR') {
-	$$rref = &$sub(%args);
-    } elsif (ref($rref) eq 'ARRAY') {
-	@$rref = &$sub(%args);
-    } else {
-	&$sub(%args);
-    }
-    $INTERP->{stack}->[0] = {%saveLocals};
-    $$cref = $content if ref($cref) eq 'SCALAR';
-
-    return 1;
-}
-
 sub mc_call_stack ()
 {
-    die "mc_call_stack $no_interp_error" if !$INTERP;
     return map($_->{truePath},@{$INTERP->{stack}});
 }
 
 sub mc_comp
 {
-    die "mc_comp $no_interp_error" if !$INTERP;
     my ($compPath, %args) = @_;
 
     $compPath = &$process_comp_path($compPath);
@@ -217,7 +147,6 @@ sub mc_comp
 
 sub mc_comp_exists
 {
-    die "mc_comp_exists $no_interp_error" if !$INTERP;
     my ($compPath, %args) = @_;
 
     $compPath = &$process_comp_path($compPath);
@@ -240,42 +169,31 @@ sub mc_comp_exists
 
 sub mc_comp_source
 {
-    die "mc_comp_source $no_interp_error" if !$INTERP;
     my ($compPath) = @_;
     
     $compPath = &$process_comp_path($compPath);
     return $INTERP->comp_root.$compPath;
 }
 
-sub mc_comp_stack ()
-{
-    die "mc_comp_stack $no_interp_error" if !$INTERP;
-    return map($_->{truePath},@{$INTERP->{stack}});
-}
-
 #
 # Version of DateManip::UnixDate that uses interpreter's notion
 # of current time and caches daily results.
 #
 sub mc_date ($)
 {
-    die "mc_date $no_interp_error" if !$INTERP;
     my ($format) = @_;
 
     my $time = $INTERP->current_time();
     if ($format =~ /%[^yYmfbhBUWjdevaAwEDxQF]/ || $time ne 'real' || !defined($INTERP->data_cache_dir)) {
-	if ($time eq 'real') {
-	    return Date::Manip::UnixDate('now',$format);
-	} else {
-	    return Date::Manip::UnixDate("epoch $time",$format);
-	}
+	$time = 'now' if $time eq 'real';
+	return UnixDate($time,$format);
     } else {
 	my %cacheOptions = (cache_file=>($INTERP->data_cache_filename('_global')),key=>'mc_date_formats',memory_cache=>($INTERP->{data_cache_store}));
 	my $href = HTML::Mason::Utils::access_data_cache(%cacheOptions);
 	if (!$href) {
 	    my %dateFormats;
 	    my @formatChars = qw(y Y m f b h B U W j d e v a A w E D x Q F);
-	    my @formatVals = split("\cA",Date::Manip::UnixDate('now',join("\cA",map("%$_",@formatChars))));
+	    my @formatVals = split("\cA",UnixDate('now',join("\cA",map("%$_",@formatChars))));
 	    my $i;
 	    for ($i=0; $i<@formatChars; $i++) {
 		$dateFormats{$formatChars[$i]} = $formatVals[$i];
@@ -290,45 +208,103 @@ sub mc_date ($)
 
 sub mc_file ($)
 {
-    die "mc_file $no_interp_error" if !$INTERP;
     my ($file) = @_;
-    # filenames beginning with / or a drive letter (e.g. C:/) are absolute
-    unless ($file =~ /^([A-Za-z]:)?\//) {
-	if ($INTERP->static_file_root) {
-	    $file = $INTERP->static_file_root . "/" . $file;
-	} else {
-	    $file = $INTERP->comp_root . (chop_slash($INTERP->locals->{parentPath})) . "/" . $file;
-	}
+    if (substr($file,0,1) ne '/') {
+	$file = $INTERP->static_file_root . "/" . $file;
     }
-    $INTERP->call_hooks('start_file',$file);
+    $INTERP->call_hooks(type=>'start_file',params=>[$file]);
     my $content = read_file($file);
-    $INTERP->call_hooks('end_file',$file);
+    $INTERP->call_hooks(type=>'end_file',params=>[$file]);
     return $content;
 }
 
 sub mc_file_root ()
 {
-    die "mc_file_root $no_interp_error" if !$INTERP;
     return $INTERP->static_file_root;
 }
 
+sub mc_form_hidden (%)
+{
+    my (%args) = @_;
+    my $hidden = '<input type="hidden" name="%s" value="%s">'."\n";
+    my $fstr;
+    while (my ($key,$value) = each(%args)) {
+	if (!ref($value)) {
+	    $fstr .= sprintf($hidden,$key,$value);
+	} elsif (ref($value) eq 'ARRAY') {
+	    foreach (@$value) {
+		$fstr .= sprintf($hidden,$key,$_);
+	    }
+	} elsif (ref($value) eq 'HASH') {
+	    my %h = %$value;
+	    foreach (keys(%h)) {
+		$fstr .= sprintf($hidden,$key,$_);
+		$fstr .= sprintf($hidden,$key,$h{$_});
+	    }
+	} else {
+	    die "mc_hidden_inputs: cannot pass ".ref($value)." reference in option string\n";
+	}
+    }
+    return $fstr;
+}
+
+sub mc_hlink ($%)
+{
+    my ($url, %options) = @_;
+    my $ostr = &$construct_option_string(%options);
+    return $ostr ? "$url?$ostr" : $url;
+}
+
 sub mc_out ($)
 {
-    die "mc_out $no_interp_error" if !$INTERP;
     $INTERP->locals->{sink}->($_[0]);
 }
 
+sub mc_pack
+{
+    my ($type,@args) = @_;
+    die "mc_pack: unknown thunk type '$type'\n" if ($type !~ /(text|file|comp|code)$/);
+    die "mc_pack: not enough arguments\n" if !@args;
+    return [$type,@args];
+}
+
+sub mc_suppress_hooks ($)
+{
+    foreach my $hookname (@_) {
+	$INTERP->suppress_hooks(name=>$hookname);
+    }
+}
+
 sub mc_time
 {
-    die "mc_time $no_interp_error" if !$INTERP;
     my $time = $INTERP->current_time;
     $time = time() if $time eq 'real';
     return $time;
 }
 
+sub mc_unpack
+{
+    my ($thunk) = @_;
+    die "mc_unpack: not a proper thunk\n" if (ref($thunk) ne 'ARRAY');
+    my ($type,@args) = @$thunk;
+    die "mc_unpack: not enough data in thunk\n" if !defined(@args);
+    if ($type eq 'text') {
+	return $args[0];
+    } elsif ($type eq 'file') {
+	return mc_file($args[0]);
+    } elsif ($type eq 'comp') {
+	return mc_comp(@args);
+    } elsif ($type eq 'code') {
+	my $sub = $args[0];
+	die "mc_unpack: argument for code type is not a code reference\n" if ref($sub) ne 'CODE';
+	return &$sub();
+    } else {
+	die "mc_unpack: unknown thunk type '$type'\n";
+    }
+}
+
 sub mc_var ($)
 {
-    die "mc_var $no_interp_error" if !$INTERP;
     my ($field) = @_;
     return $INTERP->{vars}->{$field};
 }
