@@ -12,28 +12,19 @@ require Exporter;
 use strict;
 use File::Path;
 use File::Basename;
-use File::Find;
+use File::Recurse;
+use Safe;
 use HTML::Mason::Tools qw(read_file);
 use vars qw($AUTOLOAD);
 
 my %fields =
     (preamble => '',
      postamble => '',
-     preprocess => undef,
-     postprocess => undef,
      use_strict => 1,
      source_refer_predicate => sub { return ($_[1] >= 5000) },
-     ignore_warnings_expr => 'Subroutine .* redefined',
-     taint_check => 0,
-     in_package => 'HTML::Mason::Commands',
-     allow_globals => []
+     ignore_warnings_expr => undef,
+     safe_compartment => undef
      );
-# Minor speedup: create anon. subs to reduce AUTOLOAD calls
-foreach my $f (keys %fields) {
-    next if $f =~ /^allow_globals$/;  # don't overwrite real sub.
-    no strict 'refs';
-    *{$f} = sub {my $s=shift; return @_ ? ($s->{$f}=shift) : $s->{$f}};
-}
 
 sub new
 {
@@ -47,22 +38,21 @@ sub new
 	if (exists($fields{$key})) {
 	    $self->{$key} = $value;
 	} else {
-	    die "HTML::Mason::Parser::new: invalid option '$key'\n";
+	    die "HTML::MasonParser::new: invalid option '$key'\n";
 	}
     }
     bless $self, $class;
     return $self;
 }
 
-sub allow_globals
+sub safe_compartment
 {
-    my ($self, @decls) = @_;
-    if (my @bad = grep(!/^[\$@%]/,@decls)) {
-	die "allow_globals: bad parameter '$bad[0]', must begin with one of $, @, %\n";
+    my ($self, $cpt) = @_;
+    if (defined($cpt)) {
+	$cpt->share_from('HTML::Mason::Commands',[qw($INTERP mc_comp mc_file)]);
+	$self->{safe_compartment} = $cpt;
     }
-    my %h = map {$_=>1} (@{$self->{'allow_globals'}},@decls);
-    $self->{'allow_globals'} = [keys(%h)];
-    return @{$self->{'allow_globals'}};
+    return $self->{safe_compartment};
 }
 
 sub parse
@@ -75,9 +65,8 @@ sub parse
     my $pureTextFlagRef = $options{pure_text_flag};
     my $errorRef = $options{error};
     my $wrapErrors = $options{wrap_errors};
-    my $saveTo = $options{save_to};
     my ($sub, $err, $errpos);
-    my $pureTextFlag = ($self->{preprocess} || $self->{postprocess}) ? 0 : 1;
+    my $pureTextFlag = 1;
     my $parseError = 1;
 
     #
@@ -89,42 +78,57 @@ sub parse
     }
 
     #
-    # Preprocess the script.  The preprocessor routine is handed a
-    # reference to the entire script.
+    # From here on in we must make sure not to alter the number of
+    # characters in the source, or else we will mess up source
+    # references!
     #
-    if ($self->{preprocess}) {
-        eval {$self->{preprocess}->(\$script)};
-        if ($@) {
-            $err = "error during custom preprocess step:\n$@";
-            goto parse_error;
-        }
-    }
+    # Clean up ctrl chars
+    #
+    $script =~ s/[\cM\cA\cB]/ /g;
+
+    #
+    # Pre-substitute \cA and \cB for %PERL and /%PERL to ease parsing.
+    #
+    my $perlbeginmark = "\cA"x7;
+    my $perlendmark = "\cB"x8;
+    $script =~ s@<%PERL>@$perlbeginmark@gei;
+    $script =~ s@</%PERL>@$perlendmark@gei;
     
     #
-    # Extract special sections. For backward compatibility, allow
-    # names to be prefixed with perl_.  Record all other text ranges
-    # in @textsegs.
+    # Extract perl_init, perl_cleanup, perl_args, and documentation sections.
+    # Record all other text ranges in @textsegs.
     #
-    my %sectiontext = (map(($_,''),qw(args cleanup doc filter init once)));
+    my %sectiontext = (perl_init=>'',perl_cleanup=>'',perl_args=>'',perl_doc=>'');
     my $curpos = 0;
     my @textsegs;
     my $scriptlength = length($script);
-    while ($script =~ /(<%(?:perl_)?(args|cleanup|doc|filter|init|once|text)>)/ig) {
-	my ($begintag,$beginfield) = ($1,lc($2));
-	my $beginmark = pos($script)-length($begintag);
-	my $begintail = pos($script);
-	push(@textsegs,[$curpos,$beginmark-$curpos]);
-	if ($script =~ m/(<\/%(?:perl_)?$beginfield>\n?)/ig) {
-	    my $endtag = $1;
-	    my $endmark = pos($script)-length($endtag);
-	    if ($beginfield eq 'text') {
-		# Special case for <%text> sections: add a special
-		# segment that won't get parsed
-		push(@textsegs,[$begintail,$endmark-$begintail,'<%text>']);
-	    } else {
-		$sectiontext{lc($beginfield)} .= substr($script,$begintail,$endmark-$begintail);
+    while ($script =~ /<%perl_/ig) {
+	my $beginmark = pos($script)-7;
+	my $begintail = index($script,'>',$beginmark+7);
+	next if $begintail==-1;
+	my $beginfield = lc(substr($script,$beginmark+2,$begintail-($beginmark+2)));
+	next if ($beginfield !~ /^(perl_init|perl_cleanup|perl_args|perl_doc)$/);
+
+	pos($script) = $begintail;
+	if ($script =~ m@</%perl_@ig) {
+	    my $endmark = pos($script)-8;
+	    my $endtail = index($script,'>',$endmark+8);
+	    if ($endtail==-1) {
+		$err = "<%$beginfield> with no matching </%$beginfield>";
+		$errpos = $beginmark;
+		goto parse_error;
 	    }
-	    $curpos = pos($script);
+	    my $endfield = lc(substr($script,$endmark+3,$endtail-($endmark+3)));
+	    if ($endfield ne $beginfield) {
+		$err = "section begins with <%$beginfield> and ends with </%$endfield>";
+		$errpos = $beginmark;
+		goto parse_error;
+	    }
+	    $sectiontext{$beginfield} .= substr($script,$begintail+1,($endmark-($begintail+1)));
+	    push(@textsegs,[$curpos,$beginmark-$curpos]);
+	    $curpos = $endtail+1;
+	    $curpos++ if (substr($script,$curpos,1) eq "\n");
+	    pos($script) = $curpos;
 	    $pureTextFlag = 0;
 	} else {
 	    $err = "<%$beginfield> with no matching </%$beginfield>";
@@ -144,9 +148,9 @@ sub parse
     #
     # Process args section.
     #
-    if ($sectiontext{args}) {
+    if ($sectiontext{perl_args}) {
 	my (@vars);
-	my @decls = split("\n",$sectiontext{args});
+	my @decls = split("\n",$sectiontext{perl_args});
 	@decls = grep(/\S/,@decls);
 	my $argsec = "\nmy (\$val);\n";
 	foreach my $decl (@decls) {
@@ -202,19 +206,18 @@ sub parse
 	#
 	# Declare the args as lexically scoped locals.
 	#
-	if (@vars) {
-	    $body .= "my (".join(",",@vars).");\n{".$argsec."}\n";
-	}
+	$body .= "my (".join(",",@vars).");\n{".$argsec."}\n";
     }
-
+    
     #
-    # Parse in-line insertions, which take one of four forms:
+    # Parse perl insertions.
+    # Perl insertions take one of three forms:
     #   - Lines beginning with %
-    #   - Text delimited by <%perl> </%perl>
-    #   - Text delimited by <% %> 
-    #   - Text delimited by <& &>
+    #   - Text delimited by <%PERL> </%PERL>
+    #   - Text delimited by <% %> - evaluated and output
     # All else is a string to be delimited by single quotes and output.
     #
+    # Pre-substituted \cA and \cB for %PERL to ease parsing.
     # $startline keeps track of whether the first character is the
     # start of a line.
     # $curpos keeps track of where we are in $text.
@@ -223,14 +226,7 @@ sub parse
     #
     my $startline = 1;
     my (@alphasecs, @perltexts);
-
     foreach my $textseg (@textsegs) {
-	# Special case for <%text> sections
-	if (@{$textseg}>2 && $textseg->[2] eq '<%text>') {
-	    push(@alphasecs,[$textseg->[0],$textseg->[1]]);
-	    push(@perltexts,'');
-	    next;
-	}
 	my $segbegin = $textseg->[0];
 	my $textlength = $textseg->[1];
 	my $text = substr($script,$segbegin,$textlength);
@@ -250,12 +246,10 @@ sub parse
 	    }
 	    $startline=0;
 	    my $a = index($text,"\n%",$curpos);
-	    my $b = index($text,"<%",$curpos);
-	    my $c = index($text,"<&",$curpos);
+	    my $b = index($text,$perlbeginmark,$curpos);
+	    my $c = index($text,"<%",$curpos);
 	    if ($a>-1 && ($b==-1 || $a<$b) && ($c==-1 || $a<$c)) {
-		#
 		# Line beginning with %
-		#
 		my $beginline = $a+1;
 		$alpha = [$curpos,$beginline-$curpos];
 		my $endline = index($text,"\n",$beginline);
@@ -265,69 +259,38 @@ sub parse
 		$curpos = $endline+1;
 		$startline = 1;
 		$pureTextFlag = 0;
-	    } elsif ($b>-1 && ($c==-1 || $b<$c)) {
-		#
-		# Tag beginning with <%
-		#
-		if (lc(substr($text,$b,7)) eq '<%perl>') {
-		    #
-		    # <%perl> section
-		    #
-		    $alpha = [$curpos,$b-$curpos];
-		    if (!($text =~ m{</%perl>}ig)) {
-			$err = "<%PERL> with no matching </%PERL>";
-			$errpos = $segbegin + $b;
-			goto parse_error;
-		    }
-		    my $i = pos($text)-8;
-		    my $length = $i-($b+7);
-		    $perl = substr($text,$b+7,$length);
-		    $curpos = $b+7+$length+8;
-		} else {
-		    #
-		    # <% %> section
-		    #
-		    $alpha = [$curpos,$b-$curpos];
-		    # See if this is a mistaken <%xxx> command
-		    if (substr($text,$b+2,20) =~ /^(\w+)>/) {
-			$err = "unknown section <%$1>";
-			$errpos = $segbegin + $b;
-			goto parse_error;
-		    }
-		    my $i = index($text,"%>",$b+2);
-		    if ($i==-1) {
-			$err = "'<%' with no matching '%>'";
-			$errpos = $segbegin + $b;
-			goto parse_error;
-		    }
-		    my $length = $i-($b+2);
-		    $perl = '$_out->('.substr($text,$b+2,$length).');';
-		    $curpos = $b+2+$length+2;
-		}
-		$pureTextFlag = 0;
-	    } elsif ($c>-1) {
-		#
-		# <& &> section
-		#
-		$alpha = [$curpos,$c-$curpos];
-		my $i = index($text,"&>",$c+2);
+	    } elsif ($b>-1 && ($a==-1 || $b<$a) && ($c==-1 || $b<$c)) {
+		# Text delimited by <%PERL> </%PERL>
+		$alpha = [$curpos,$b-$curpos];
+		my $i = index($text,$perlendmark,$b+7);
 		if ($i==-1) {
-		    $err = "'<&' with no matching '&>'";
+		    $err = "<%PERL> with no matching </%PERL>";
+		    $errpos = $segbegin + $b;
+		    goto parse_error;
+		}
+		my $length = $i-($b+7);
+		$perl = substr($text,$b+7,$length);
+		$curpos = $b+7+$length+8;
+		$pureTextFlag = 0;
+	    } elsif ($c>-1 && ($a==-1 || $c<$a) && ($b==-1 || $c<$b)) {
+		# Text delimited by <% %>
+		$alpha = [$curpos,$c-$curpos];
+		# See if this is a mistaken <%xxx> command
+		if (substr($text,$c+2,20) =~ /^(\w+)>/) {
+		    $err = "unknown section <%$1>";
+		    $errpos = $segbegin + $c;
+		    goto parse_error;
+		}
+		my $i = index($text,"%>",$c+2);
+		if ($i==-1) {
+		    $err = "'<%' with no matching '%>'";
 		    $errpos = $segbegin + $c;
 		    goto parse_error;
 		}
 		my $length = $i-($c+2);
-		my $call = substr($text,$c+2,$length);
-		for ($call) { s/^\s+//; s/\s+$// }
-		if (substr($call,0,1) =~ /[A-Za-z0-9\/_.]/) {
-		    # Literal component path; put quotes around it
-		    my $comma = index($call,',');
-		    $comma = length($call) if ($comma==-1);
-		    (my $comp = substr($call,0,$comma)) =~ s/\s+$//;
-		    $call = "'$comp'".substr($call,$comma);
-		}
-		$perl = "mc_comp($call);";
+		$perl = '$_out->('.substr($text,$c+2,$length).');';
 		$curpos = $c+2+$length+2;
+		#$curpos++ if (substr($text,$curpos,1) eq "\n");
 		$pureTextFlag = 0;
 	    } else {
 		# No more special characters, take the rest.
@@ -355,14 +318,14 @@ sub parse
     #
     # Determine length of plain text.
     #
-    my $alphalength=0;
+    my $alphalength;
     foreach (@alphasecs) {$alphalength += $_->[1]}
 
     #
     # Use source_refer_predicate to determine whether to use source
     # references or directly embedded text.
     #
-    my $useSourceReference = $self->source_refer_predicate->($scriptlength,$alphalength) && !$self->{preprocess} && !$self->{postprocess};
+    my $useSourceReference = $self->source_refer_predicate->($scriptlength,$alphalength);
     my @alphatexts;
     if ($useSourceReference) {
 	$body .= 'my $_srctext = mc_file($INTERP->locals->{sourceFile});'."\n";
@@ -370,7 +333,7 @@ sub parse
     } else {
 	foreach (@alphasecs) {
 	    my $alpha = substr($script,$_->[0],$_->[1]);
-	    $alpha =~ s{([\\\'])} {\\$1}g;        # escape backslashes and single quotes
+	    $alpha =~ s/\'/\\\'/g;
 	    push(@alphatexts,sprintf('$_out->(\'%s\');',$alpha));
 	}
     }
@@ -381,41 +344,15 @@ sub parse
     $body .= '$INTERP->debug_hook($INTERP->locals->{truePath}) if (%DB::);'."\n";
     
     #
-    # Insert <%filter> section.
+    # Insert begin section.
     #
-    if ($sectiontext{filter}) {
-	my $ftext = $sectiontext{filter};
-	for ($ftext) { s/^\s+//g; s/\s+$//g }
-	$body .= sprintf('{ my ($_c,$_r); if (mc_call_self(\$_c,\$_r)) { for ($_c) { %s } mc_out($_c); return $_r }};',$ftext);
-    }
-
-    #
-    # Insert <%init> section.
-    #
-    $body .= $sectiontext{init}."\n";
+    $body .= $sectiontext{perl_init}."\n";
 
     #
     # Call start_primary hooks.
     #
-    $body .= "\$INTERP->call_hooks('start_primary');\n";
+    $body .= "\$INTERP->call_hooks(type=>'start_primary');\n";
     
-    #
-    # Postprocess the alphabetical and Perl stuff separately
-    #
-    if ($self->{postprocess})
-    {
-        foreach my $a (@alphatexts)
-        {
-            next unless $a;
-            $self->{postprocess}->(\$a, 'alpha');
-        }
-        foreach my $p (@perltexts)
-        {
-            next unless $p;
-            $self->{postprocess}->(\$p, 'perl');
-        }
-    }
-
     #
     # Append text and perl sections to body.
     #
@@ -429,12 +366,12 @@ sub parse
     #
     # Call end_primary hooks.
     #
-    $body .= "\$INTERP->call_hooks('end_primary');\n";
+    $body .= "\$INTERP->call_hooks(type=>'end_primary');\n";
     
     #
-    # Insert <%cleanup> section.
+    # Insert end section.
     #
-    $body .= $sectiontext{cleanup}."\n";
+    $body .= $sectiontext{perl_cleanup}."\n";
 
     #
     # Insert user postamble and return undef by default.
@@ -443,21 +380,16 @@ sub parse
     $body .= "return undef;\n";
 
     #
-    # Wrap body in subroutine and add header, including <%once> section.
+    # Wrap body in subroutine and add header.
     #
     my $header = "";
-    my $pkg = $self->{in_package};
-    $header .= "package $pkg;\n";
-    $header .= "use strict;\n" if $self->use_strict;
-    $header .= sprintf("use vars qw(%s);\n",join(" ","\$INTERP",@{$self->{'allow_globals'}}));
-    $header .= "\n".$sectiontext{once}."\n" if $sectiontext{once};
-    $body = "$header\nsub {\n$body\n};\n";
+    if (!$self->safe_compartment) {
+	$header .= "package HTML::Mason::Commands;\n";
+	$header .= "use strict;\n" if $self->use_strict;
+	$header .= "use vars qw(\$INTERP);\n";
+    }
+    $body = "$header\nreturn sub {\n$body\n};\n";
 
-    #
-    # Eliminate DOS ctrl-M chars
-    #
-    $body =~ s/\cM//g;
-    
     #
     # Check for errors and warnings.
     #
@@ -475,7 +407,7 @@ sub parse
 	    $err =~ s/\'/\\\'/g;
 	    $err =~ s/\(eval [0-9]\) //g;
 	    my $msg = sprintf("Error during compilation%s:",$scriptFile ? " of '$scriptFile'" : "");
-	    my $errscript = "$header\nsub {\ndie \"$msg\\n\".'$err'.\"\\n\"\n}\n";
+	    my $errscript = "$header\nreturn sub {\ndie \"$msg\\n\".'$err'.\"\\n\"\n}\n";
 	    $sub = eval($errscript);
 	} else {
 	    $err .= "\n";
@@ -483,15 +415,8 @@ sub parse
 	    $$errorRef = $err if defined($errorRef);
 	}
     }
-    
-    if (!$err && defined($saveTo)) {
-	File::Path::mkpath(File::Basename::dirname($saveTo));
-	my $fh = new IO::File ">$saveTo" or die "Couldn't write object file $saveTo";
-	print $fh $body if (!$pureTextFlag);
-	$fh->close;
-    }
-    $$resultTextRef = $body if !$parseError && defined($resultTextRef);
     $$resultCodeRef = $sub if $success && defined($resultCodeRef);
+    $$resultTextRef = $body if !$parseError && defined($resultTextRef);
     $$pureTextFlagRef = $pureTextFlag if defined($pureTextFlagRef);
     return $success;
 }
@@ -504,15 +429,21 @@ sub evaluate
     {
 	my $warnstr;
 	local $^W = 1;
-	local $SIG{__WARN__} = $ignoreExpr ? sub { $warnstr .= $_[0] if $_[0] !~ /$ignoreExpr/ } : sub { $warnstr .= $_[0] };
+	local $SIG{__WARN__} = $ignoreExpr ? sub { $warnstr .= $_[0] if $_[0] =~ /$ignoreExpr/ } : sub { $warnstr .= $_[0] };
+	my $cpt = $self->safe_compartment;
 	if ($options{script}) {
-	    my $script = $options{script};
-	    ($script) = ($script =~ /^(.*)$/s) if $self->taint_check;
-	    $sub = eval($script);
+	    if (!$cpt) {
+		$sub = eval($options{script});
+	    } else {
+		$sub = $cpt->reval($options{script});
+	    }
 	} elsif ($options{script_file}) {
-	    my $file = $options{script_file};
-	    ($file) = ($file =~ /^(.*)$/s) if $self->taint_check;
-	    $sub = do($file);
+	    die "evaluate: script_file '$options{script_file}' not a regular file" if (!-f $options{script_file});
+	    if (!$cpt) {
+		$sub = do($options{script_file});
+	    } else {
+		$sub = $cpt->rdo($options{script});
+	    }
 	} else {
 	    die "evaluate: must specify script or script_file";
 	}
@@ -535,18 +466,17 @@ sub evaluate
 
 sub make {
     my ($self, %options) = @_;
-    my $compRoot = $options{comp_root} or die "make: must specify comp_root\n";
-    my $dataDir = $options{data_dir} or die "make: must specify data_dir\n";
-    die "make: source_dir '$compRoot' does not exist\n" if (!-d $compRoot);
-    die "make: object_dir '$dataDir' does not exist\n" if (!-d $dataDir);
-    my $sourceDir = $compRoot;
-    my $objectDir = "$dataDir/obj";
-    my $errorDir = "$dataDir/errors";
+    my $sourceDir = $options{source_dir} or die "make: must specify source_dir\n";
+    my $objectDir = $options{object_dir} or die "make: must specify object_dir\n";
+    my $errorDir = $options{error_dir} or die "make: must specify error_dir\n";
+    die "make: object_dir must be different from source_dir\n" if $sourceDir eq $objectDir;
+    die "make: source_dir '$sourceDir' does not exist\n" if (!-d $sourceDir);
+    die "make: object_dir '$objectDir' does not exist\n" if (!-d $objectDir);
     my @paths = (exists($options{paths})) ? @{$options{paths}} : ('/');
     my $verbose = (exists($options{verbose})) ? $options{verbose} : 1;
     my $predicate = $options{predicate} || sub { $_[0] !~ /\~/ };
     my $dirCreateMode = $options{dir_create_mode} || 0775;
-    my $reloadFile = $options{update_reload_file} ? "$dataDir/etc/reload.lst" : undef;
+    my $reloadFile = $options{reload_file};
     my ($relfh);
     if (defined($reloadFile)) {
 	$relfh = new IO::File ">>$reloadFile" or die "make: cannot open '$reloadFile' for writing\n";
@@ -554,7 +484,7 @@ sub make {
     }
     
     my $compilesub = sub {
-	my ($srcfile) = $File::Find::name;
+	my ($srcfile) = @_;
 	return if (!-f $srcfile);
 	return if defined($predicate) && !($predicate->($srcfile));
 	my $compPath = substr($srcfile,length($sourceDir));
@@ -608,8 +538,7 @@ sub make {
 	if (-f $fullpath) {
 	    $compilesub->($fullpath);
 	} elsif (-d $fullpath) {
-	    my $sub = sub {$compilesub->($_) if -f};
-	    find($sub,$fullpath);
+	    recurse {$compilesub->($_)} $fullpath;
 	} else {
 	    die "make: no such file or directory '$fullpath'\n";
 	}
@@ -624,7 +553,15 @@ sub AUTOLOAD {
     $name =~ s/.*://;   # strip fully-qualified portion
     return if $name eq 'DESTROY';
 
-    die "No such function `$name' in class $type";
+    unless (exists $self->{'_permitted'}->{$name} ) {
+        die "No such function `$name' in class $type";
+    }
+
+    if (@_) {
+        return $self->{$name} = shift;
+    } else {
+        return $self->{$name};
+    }
 }
 1;
 
