@@ -74,6 +74,12 @@ sub cgi_env {
     return %env;
 }
 
+sub cgi_var {
+    my ($self,$key) = @_;
+    my $value = exists($self->{env}->{$key}) ? $self->{env}->{$key} : $self->{r}->cgi_var($key);
+    return $value;
+}
+
 sub dir_config {
     my ($self,$key) = @_;
     return $self->{dir_config}->{$key};
@@ -84,8 +90,8 @@ sub AUTOLOAD {
     my $name = $AUTOLOAD;
     $name =~ s/.*://;   # strip fully-qualified portion
     return if $name eq 'DESTROY';
-    my $realr = $self->{r};
-    eval(sprintf('$realr->%s(@_)',$name));
+    my $r = $self->{r};
+    eval(sprintf('$r->%s(@_)',$name));
 }
 
 #
@@ -107,12 +113,11 @@ sub send_http_header {
 }
 
 package HTML::Mason::Preview;
+use CGI::Base;
 use HTML::Mason::ApacheHandler;
-use HTML::Mason::Config;
-use HTML::Mason::Tools qw(date_delta_to_secs html_escape);
-use MLDBM ($HTML::Mason::Config{mldbm_use_db}, $HTML::Mason::Config{mldbm_serializer});
 use IO::File qw(!/^SEEK/);
 use POSIX;
+use Date::Manip;
 
 sub open_preview_settings
 {
@@ -140,7 +145,7 @@ sub handle_preview_request
 	$r->content_type('text/html');
 	$r->header_out(Pragma=>'no-cache');
 	$r->send_http_header();
-	print("<h2>Previewer Error</h2>\n<pre>\n$err\n</pre>\n");
+	$r->print("<h2>Previewer Error</h2>\n<pre>\n$err\n</pre>\n");
 	return;
     }
     return $result;
@@ -165,7 +170,7 @@ sub handle_preview_request_1
     #
     my $in = open_preview_settings($ah->preview_dir,$userName,0);
     my $href = $in->{$port};
-    die "Cannot find preview configuration for user '$userName', port $port\n. Go to the main previewer page to set the configuration for this port.\n" if (!$href);
+    die "cannot find preview configuration for user '$userName', port $port\n" if (!$href);
     my $conf = $href->{request};
 
     my $interp = $ah->interp;
@@ -191,7 +196,7 @@ sub handle_preview_request_1
 	my $time;
 	if ($timetype ne 'real') {
 	    if ($timetype eq 'relative') {
-		$time = time() + date_delta_to_secs($conf->{'time'}->{delta});
+		$time = DateCalc('now',$conf->{'time'}->{delta});
 	    } elsif ($timetype eq 'absolute') {
 		$time = $conf->{'time'}->{value};
 	    }
@@ -268,7 +273,7 @@ sub handle_preview_request_1
     }
 
     if ($conf->{output_type} eq 'HTML') {
-        return $ah->handle_request($pr);
+        $ah->handle_request($pr);
     } elsif ($conf->{output_type} eq 'Debug') {
 
 	#
@@ -314,6 +319,13 @@ sub handle_preview_request_1
 	    $content .= "\cAEVENT$eventnum\cA";
 	    $compEvents[$eventnum++] = {type=>'endFile',path=>$file,end=>$end};
 	};
+	my $readFileHook = sub {
+	    my ($self,$filePath) = @_;
+	    my ($startevent,$endevent) = ($eventnum,$eventnum+1);
+	    $compEvents[$startevent] = {type=>'startFile',path=>$filePath};
+	    $compEvents[$endevent] = {type=>'endFile',path=>$filePath};
+	    $eventnum+=2;
+	};
 
 	#
 	# Create a trace of objects (components/files), with one
@@ -324,9 +336,7 @@ sub handle_preview_request_1
 	    my ($trace,$obj);
 	    foreach $obj (@objects) {
 		next if !defined($obj);
-		$trace .= sprintf('%s<a href="%s">%s</a>  %s%s',
-				  (' ' x (4-length($obj->{label}))),
-				  $obj->{srclink},$obj->{label},$obj->{display},$obj->{repeat}>1 ? " (repeats ".$obj->{repeat}." times)": "");
+		$trace .= sprintf('%s<a href="%s">%s</a>  %s%s',(' ' x (4-length($obj->{label}))),$obj->{srclink},$obj->{label},$obj->{display});
 		$trace .= "\n";
 	    }
 	    return $trace;
@@ -341,7 +351,7 @@ sub handle_preview_request_1
 	$interp->add_hook(name=>'preview',type=>'end_file',code=>$endFileHook);
 	$interp->out_method(sub { $content .= $_[0] });
 	$ah->output_mode(undef);
-        my $statuscode = $ah->handle_request($pr);
+        $ah->handle_request($pr);
 
 	#
 	# Validate file events. For each startFile event,
@@ -383,16 +393,14 @@ sub handle_preview_request_1
 			}
 		    }
 		}
-		# leave "invalid" file events in?
-		#if (!$valid) {
-		if (0) {
+		if (!$valid) {
 		    undef($compEvents[$e]);
 		    undef($compEvents[$e+1]);
 		    $e++;
 		}
 	    }
 	}
-
+	
 	#
 	# Analyze event information. Create a list of objects
 	# (distinct component and mc_file invocations)
@@ -408,6 +416,7 @@ sub handle_preview_request_1
 		my ($objtext,$objsrclink,$objlabel,$objtype,$objdisplay);
 		$objcount++;
 		if ($type eq 'startComp') {
+		    my $compId = $event->{compId};
 		    $objtext = "$path";
 		    $objlabel = "$objcount";
 		    $objsrclink = "#object$objcount";
@@ -421,14 +430,8 @@ sub handle_preview_request_1
 		    $objsrclink = "#object$objcount";
 		    $objtype = "file";
 		}
-		my $lastobj = $objects[$objcount-1];
-		if ($objtext eq $lastobj->{text} && $objtype eq $lastobj->{type} && scalar(@stack) == $lastobj->{depth}) {
-		    $objcount--;
-		    $objects[$objcount]->{repeat}++;
-		} else {
-		    $objdisplay = ($createObjectLink) ? $createObjectLink->($objtype,$path,scalar(@stack),$objtext) : (('  ' x scalar(@stack)) . $objtext);
-		    $objects[$objcount] = {count=>$objcount,type=>$objtype,text=>$objtext,display=>$objdisplay,label=>$objlabel,color=>'003399',srclink=>$objsrclink,depth=>scalar(@stack),repeat=>1};
-		}
+		$objdisplay = ($createObjectLink) ? $createObjectLink->($objtype,$path,scalar(@stack),$objtext) : (('  ' x scalar(@stack)) . $objtext);
+		$objects[$objcount] = {count=>$objcount,text=>$objtext,display=>$objdisplay,label=>$objlabel,color=>'003399',srclink=>$objsrclink,depth=>scalar(@stack)};
 		push(@stack,$objcount);
 	    } elsif ($type =~ /^(endComp|endFile)$/) {
 		pop(@stack);
@@ -439,50 +442,41 @@ sub handle_preview_request_1
 	#
 	# Initial component trace
 	#
-	if (@objects) {
-	    $trace = &$createTraceSub(@objects);
-	    $pr->print("<h2>Component trace</h2>\n<pre>\n$trace</pre>\n");
-	}
+	$trace = &$createTraceSub(@objects);
+	$pr->print("<h2>Component trace</h2>\n<pre>\n$trace</pre>\n");
 
 	#
 	# Content
 	#
-	if ($content) {
-	    $pr->print("<h2>Content</h2>\n<pre>\n");
-	    $content =~ s/\cM//g;
-	    ($content) = html_escape($content);
-	    $content =~ s/\cAEVENT/\n\cAEVENT/g;
-	    my ($obj);
-	    while ($content =~ /(.*)/g) {
-		my $line = $1;
-		my $beginFlag = 0;
-		if ($line =~ /\cAEVENT([0-9]+)\cA/) {
-		    $eventnum = $1;
-		    my $event = $compEvents[$eventnum];
-		    $line =~ s/\cAEVENT[0-9]+\cA//;
-		    if (defined($event)) {
-			if ($event->{type} =~ /^end/ && $obj) {
-			    $pr->print(sprintf("<font color=#%s>%5s &lt;&lt;</font>\n",$obj->{color},$obj->{label}));
-			}
-			$obj = $objects[$event->{objnum}];
-			if ($event->{type} =~ /^start/) {
-			    $pr->print("<a name=object$obj->{count}>");
-			    $beginFlag = 1;
-			}
+	$pr->print("<h2>Content</h2>\n<pre>\n");
+	$content =~ s/\cM//g;
+	($content) = html_escape($content);
+	$content =~ s/\cAEVENT/\n\cAEVENT/g;
+	my ($obj);
+	while ($content =~ /(.*)/g) {
+	    my $line = $1;
+	    my $beginFlag = 0;
+	    if ($line =~ /\cAEVENT([0-9]+)\cA/) {
+		$eventnum = $1;
+		my $event = $compEvents[$eventnum];
+		$line =~ s/\cAEVENT[0-9]+\cA//;
+		if (defined($event)) {
+		    if ($event->{type} =~ /^end/ && $obj) {
+			$pr->print(sprintf("<font color=#%s>%5s &lt;&lt;</font>\n",$obj->{color},$obj->{label}));
+		    }
+		    $obj = $objects[$event->{objnum}];
+		    if ($event->{type} =~ /^start/) {
+			$pr->print("<a name=object$obj->{count}>");
+			$beginFlag = 1;
 		    }
 		}
-		next if ($line !~ /\S/ && !$obj);
-		my $depth = 0;
-		$pr->print(sprintf("<font color=#%s>%s%5s %s</font>%s%s\n",$obj->{color},(' ' x $depth),$obj->{label},($beginFlag ? '&gt;&gt;' : '  '),(' ' x (4-$depth)),$line));
 	    }
-	    $pr->print("\n</pre>\n");
+	    next if ($line !~ /\S/ && !$obj);
+	    my $depth = 0;
+	    $pr->print(sprintf("<font color=#%s>%s%5s %s</font>%s%s\n",$obj->{color},(' ' x $depth),$obj->{label},($beginFlag ? '&gt;&gt;' : '  '),(' ' x (4-$depth)),$line));
 	}
+	$pr->print("\n</pre>\n");
 
-	#
-	# Status
-	#
-	$pr->print("<h2>Status: $statuscode</h2>\n") if $statuscode && $statuscode != 200;
-	
         #
 	# Headers in
 	#
@@ -502,9 +496,11 @@ sub handle_preview_request_1
 	#
     	$pr->print("<h2>Headers out</h2>\n<ul>");
 	my %outHeaders = $r->headers_out();
-	my $gmtdate = gmtime;
+	my $gmtdate;
+	{ local $Date::Manip::ConvTZ = 'GMT'; $gmtdate = UnixDate('now','%g') }
+	$gmtdate =~ s@,(\S)@, $1@;
 	$pr->print("<li type=circle><b>Date:</b>  $gmtdate\n");
-	my $serverType = $ENV{SERVER_SOFTWARE} || 'Apache/1.2 mod_perl/1.0'; 
+	my $serverType = $r->cgi_var('SERVER_SOFTWARE') || 'Apache/1.2 mod_perl/1.0'; 
 	$pr->print("<li type=circle><b>Server:</b>  $serverType\n");
 	while (($key,$value) = each(%outHeaders)) {
 	    $pr->print("<li type=disc><b>$key:</b>  $value\n");
